package gatewayapi.sample;

import java.util.Collection;
import java.util.Optional;
import java.util.Scanner;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.mmbnetworks.gatewayapi.CommandData;
import com.mmbnetworks.gatewayapi.ConnectionInfo;
import com.mmbnetworks.gatewayapi.ConnectionInfo.ConnectionType;
import com.mmbnetworks.gatewayapi.GatewayClient;
import com.mmbnetworks.gatewayapi.PropertyCommandData;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.LightDevice;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.gatewayapi.entity.ThermostatDevice;
import com.mmbnetworks.gatewayapi.entity.ThermostatDevice.FanMode;
import com.mmbnetworks.gatewayapi.entity.ThermostatDevice.Setpoint;
import com.mmbnetworks.gatewayapi.entity.ThermostatDevice.SystemMode;
import com.mmbnetworks.gatewayapi.event.device.DeviceEvent;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventHandler;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventStatus;
import com.mmbnetworks.gatewayapi.exception.GatewayConnectionException;
import com.mmbnetworks.gatewayapi.exception.InvalidInputException;
import com.mmbnetworks.rapidconnectdevice.zcl.ZigBee;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.MessageType;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.ZCLMulticastMessage.ZCLMulticastMessageBuilder;

import ch.qos.logback.classic.Level;

/**
 * This sample application demonstrates how one might use the GatewayClient to connect to a known
 * network interface and interact with devices on it.
 * <p>
 * Usage: SampleApp portName
 * <p>
 * Example 1: java -jar SampleApp.jar
 * Example 2: java -jar SampleApp.jar /dev/ttyUSB0
 * Example 3: java -jar SampleApp.jar /dev/ttyUSB0 /path/to/config/file.xml
 * 
 */
public class SimpleCli implements DeviceEventHandler {

    public static void main(String[] args) {

        GatewayClient gw = null;
        Scanner scanner = new Scanner(System.in, "UTF-8");

        // if no args, prompt user to enter serial port and use default config file
        if (args.length == 0) {
            while (true) {
                logger("Enter a serial port (i.e. /dev/ttyUSB0), or 'q' to exit:");
                final String input = scanner.nextLine();

                if ("q".equalsIgnoreCase(input)) {
                    logger("Quitting...");
                    scanner.close();
                    return;

                } else {
                    try {
                        gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, input), new SimpleCli());
                        break;
                    } catch (GatewayConnectionException e) { 
                        logger("Could not connect or create GatewayClient, try again");
                    }
                }
            }
        // if one arg provided, assume it's the port and use default config file
        } else if (args.length == 1) {
            final String devicePort = args[0];
            try {
                gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, devicePort), new SimpleCli());
            } catch (GatewayConnectionException e) { 
                logger("Could not connect or create GatewayClient with port: " + devicePort);
                scanner.close();
                return;
            }
        // if two args provided, assume first arg is the port and second arg is the config file path
        } else if (args.length == 2) {
            final String devicePort = args[0];
            final String configFilePath = args[1];
            try {
                gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, devicePort), new SimpleCli(), configFilePath);
            } catch (GatewayConnectionException e) { 
                logger("Could not connect or create GatewayClient with port: " + devicePort);
                scanner.close();
                return;
            }
        } else {
            logger("Invalid number of arguments.");
            scanner.close();
            return;
        }

        // print available commands
        final String[] commands = {
                "==== General Commands ====",
                " h - help",
                " toggleLog - turn logging on/off",
                " v - get version",
                " n - network status",
                " f - form network",
                " p - permit join (30 sec)",
                " c - close network permit join window",
                " l - leave network",
                " q - quit",
                " d - list devices and their device ids",
                "==== Device Commands ====",
                " getDeviceType - get a list of device ids of a certain type",
                " getLights - get all the lighting device ids",
                " getThermostats - get all the thermostat device ids",
                " getProperties - get properties for a selected device id. Will perform a network request",
                " getCachedProperties - get last known properties for a selected device id. Will be empty if there are no last cached properties",
                " getProtocolProperties - gets the properties of the device in the format of the underlying protocol (i.e. zigbee). Will perform a network request",
                " getCachedProtocolProperties - gets the last known properties of the device in the format of the underlying protocol (i.e. zigbee). Will be empty if there are no last cached properties",
                " getProperty - get a specific property. Will perform a network request",
                " getCachedProperty - get the last known value of a specific property. Will be empty if there is no last known value. Will throw exception if property doesn't exist",
                " updateProperty - update a specific property for a selected device id",
                " bind - bind a source to a destination",
                " unbind - unbind a source to a destination",
                " enableReporting - enable reporting on properties using default parameters",
                " addSwitchHandler - map a switch device to a group id",
                " addToGroup - add a device to a group",
                " removeFromGroup - remove a device from a group",
                " kickDevice - kick a remote device off the network",
                " removePropertyListeners - remove property update listeners for the selected device",
                "==== Light Commands ====",
                " on",
                " off",
                " moveToLevel - change the dimmable light level",
                " readLevel - read the dimmable light level",
                "==== Thermostat Commands ====",
                " change mode",
                " change fan mode",
                " adjust setpoint - adjust the setpoint up or down by a specific amount",
                " change setpoint limit - change the setpoint to a specified temperature",
                " read mode",
                " read fan mode",
                " read running mode",
                " read setpoint limit",
        };
        printCommands(commands);
        final ch.qos.logback.classic.Logger rootLogger = (ch.qos.logback.classic.Logger) org.slf4j.LoggerFactory.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME);
        final Level originalLogLevel = rootLogger.getLevel();
        Level currentLogLevel = originalLogLevel;
        rootLogger.setLevel(Level.OFF);
        currentLogLevel = Level.OFF;

        // use the main thread as a CLI
        while (true) {
            logger("Enter a command: ");
            final String input = scanner.nextLine();
            
            if ("h".equalsIgnoreCase(input)) {
                printCommands(commands);

            } else if ("toggleLog".equalsIgnoreCase(input)) {
                if (currentLogLevel != Level.OFF) { 
                    rootLogger.setLevel(Level.OFF);
                    currentLogLevel = Level.OFF;
                } else {
                    rootLogger.setLevel(originalLogLevel);
                    currentLogLevel = originalLogLevel;
                }

            } else if ("v".equalsIgnoreCase(input)) {
                logger("API Version = " + GatewayClient.getApiVersion());

            } else if ("q".equalsIgnoreCase(input)) {
                break;

            } else if ("n".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.getNetworkStatus(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });   
                }

            } else if ("f".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.createNetwork(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("p".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.scanForDevices(c, 30).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("c".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.scanForDevices(c, 0).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("l".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.dissolveNetwork(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("d".equalsIgnoreCase(input)) {
                final Collection<Device> devices = gw.getDevices();
                if (devices.size() == 0) {
                    logger("No devices found on the network");

                } else {
                    int i = 1;
                    for (Device d : devices) {
                        logger(i++ + ") Device Id: " + d.getID());
                    }
                }

            } else if ("getDeviceType".equalsIgnoreCase(input)) {
                logger("Enter a device type (i.e. light, thermostat, occupancy sensor, combined interface, etc): ");
                final String typeStr = scanner.nextLine();
                final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().getType().equalsIgnoreCase(typeStr) );
                int i = 1;
                for (Device d : devices) {
                    logger(i++ + ") " + typeStr + " device id: " + d.getID());
                }

            } else if ("getLights".equalsIgnoreCase(input)) {
                final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().equals(LightDevice.DEVICE_TYPE));
                int i = 1;
                for (Device d : devices) {
                    logger(i++ + ") LightDevice Id: " + d.getID());
                }

            } else if ("getThermostats".equalsIgnoreCase(input)) {
                final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().equals(ThermostatDevice.DEVICE_TYPE));
                int i = 1;
                for (Device d : devices) {
                    logger(i++ + ") ThermostatDevice Id: " + d.getID());
                }

            } else if ("getCachedProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }

                final Collection<Property> properties = d.getCachedProperties();
                if (properties.isEmpty()) {
                    logger("No properties found");
                } else {
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }   
                }

            } else if ("getProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                final CompletableFuture<Collection<Property>> result = d.getProperties();
                logger("Retrieving properties...");
                try {
                    Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated properties");
                    e.printStackTrace();
                }
            } else if ("getCachedProtocolProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }

                final Collection<Property> properties = d.getCachedProtocolProperties();
                if (properties.isEmpty()) {
                    logger("No properties found");
                } else {
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                }
                
            } else if ("getProtocolProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                final CompletableFuture<Collection<Property>> result = d.getProtocolProperties();
                logger("Retrieving protocol properties...");
                try {
                    Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated properties: " + e.getMessage());
                    e.printStackTrace();
                }
            } else if ("getProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                
                logger("Enter a property name: ");
                final String propertyName = scanner.nextLine();
                
                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                final CompletableFuture<Property> p = device.getProperty(propertyName);
                try {
                    logger("Retrieving property \'" + propertyName +"\'");
                    final Property fProp = p.get(30, TimeUnit.SECONDS);
                    logger("name: " + fProp.getName() + ", value: " + fProp.getValue());
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated property: " + e.getMessage());
                }

            } else if ("getCachedProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();

                logger("Enter a property name: ");
                final String propertyName = scanner.nextLine();

                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                final Optional<Property> pOpt = device.getCachedProperty(propertyName);
                if (!pOpt.isPresent()) {
                    logger("Property \'" + propertyName + "\' not found");
                    continue;
                }
                Property fProp = pOpt.get();
                logger("name: " + fProp.getName() + ", value: " + fProp.getValue());
            } else if ("enableReporting".equalsIgnoreCase(input)) {
                logger("Enter a device id:");
                String idStr = scanner.nextLine();
                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                try {
                    final CompletableFuture<String> f = device.enableDefaultReporting();
                    final String result = f.get(30, TimeUnit.SECONDS);
                    logger("Status for command: " + result);
                } catch (IllegalStateException | InterruptedException | ExecutionException | TimeoutException e) {
                    logger(e.getMessage());
                }
            
            } else if ("bind".equalsIgnoreCase(input)) {
                try {
                    logger("Enter a source id:");
                    final String source = scanner.nextLine();
                    final Device sourceDevice = gw.getDevice(source);
                    logger("Enter a destination id:");
                    final String destination = scanner.nextLine();
                    final Device destinationDevice = gw.getDevice(destination);
                    logger("Enter a Property name to bind:");
                    final String property = scanner.nextLine();
                    sourceDevice.bindDevice(destinationDevice.getID(), property).get()
                        .getCommandResult()
                        .whenComplete((result, error) -> {
                            if (error != null) {
                                logger("error calling bind: " + error.getMessage());
                            }
                            logger("bind device result: " + result);
                        });

                } catch (Exception ex) {
                    logger("error: " + ex.getMessage());
                }

            } else if ("addToGroup".equalsIgnoreCase(input)) {
                
                logger("Enter a device id:");
                final String source = scanner.nextLine();
                final Device d = gw.getDevice(source);
                if (d == null) { logger("no device found"); continue; }
                
                logger("Enter a group id (i.e. 1-10):");
                final String groupIdString = scanner.nextLine();
                int groupId = 0;
                try {
                    groupId = Integer.parseInt(groupIdString);
                } catch (NumberFormatException e) {
                    logger("invalid number");
                    continue;
                }
                
                final CompletableFuture<String> response = d.addToGroup(groupId);
                response.thenAccept( msg -> logger(msg) );

            } else if ("removeFromGroup".equalsIgnoreCase(input)) {
                
                logger("Enter a device id:");
                final String source = scanner.nextLine();
                final Device d = gw.getDevice(source);
                if (d == null) { logger("no device found"); continue; }
                
                logger("Enter a group id (i.e. 1-10):");
                final String groupIdString = scanner.nextLine();
                int groupId = 0;
                try {
                    groupId = Integer.parseInt(groupIdString);
                } catch (NumberFormatException e) {
                    logger("invalid number");
                    continue;
                }
                
                final CompletableFuture<String> response = d.removeFromGroup(groupId);
                response.thenAccept( msg -> logger(msg) );

            } else if ("addSwitchHandler".equalsIgnoreCase(input)) {
                
                logger("Enter a source id (i.e. the switch):");
                final String source = scanner.nextLine();
                final Device sourceDevice = gw.getDevice(source);
                if (sourceDevice == null) { logger("no device found"); continue; }
                
                logger("Enter a Group Id to re-broadcast switch message to:");
                final String groupIdString = scanner.nextLine();
                try {
                    Integer.parseInt(groupIdString);
                } catch (NumberFormatException e) {
                    logger("invalid number");
                    continue;
                }
                final GatewayClient finalGwClient = gw;
                
                sourceDevice.addProtocolHandler( (device,jsonMsg) -> {
                    logger("Rebroadcast handler: " + jsonMsg); 
                    try {
                        final JsonParser parser = new JsonParser();
                        final JsonObject jsonObject = parser.parse(jsonMsg).getAsJsonObject();
                        final JsonObject message = jsonObject.get("message").getAsJsonObject();

                        // send multicast protocol message
                        final ZCLMulticastMessageBuilder builder = new ZCLMulticastMessageBuilder();
                        builder.setGatewayAPIVersion(GatewayClient.getApiVersion());
                        builder.setProtocolName("zigbee");
                        builder.setProtocolVersion(3);
                        builder.setMessageType(MessageType.ZCL_MULTICAST);
                        builder.setGroupID(groupIdString);
                        builder.setClusterID(message.get("clusterId").getAsString());
                        builder.setFrameControl(ZigBee.FrameConstants.FRAME_CONTROL_CLIENT_TO_SERVER_CLUSTER_CMD.toBitmap8());
                        builder.setCommandID(message.get("commandId").getAsString());
                        // just send on the first available network interface; we're assuming we're only running on one for now
                        finalGwClient.sendProtocolMessage(
                                finalGwClient.getConnectionInfo().iterator().next(),
                                builder.build().toJson());

                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                });
                
            } else if ("unbind".equalsIgnoreCase(input)) {
                try {
                    logger("Enter a source id:");
                    final String source = scanner.nextLine();
                    final Device sourceDevice = gw.getDevice(source);
                    logger("Enter a destination id:");
                    final String destination = scanner.nextLine();
                    final Device destinationDevice = gw.getDevice(destination);
                    logger("Enter a Property name to unbind:");
                    final String property = scanner.nextLine();
                    sourceDevice.unbindDevice(destinationDevice.getID(), property).get()
                        .getCommandResult()
                        .whenComplete((result, error) -> {
                            if (error != null) {
                                logger("error calling unbind: " + error.getMessage());
                            }
                            logger("unbind device result: " + result);
                        });

                } catch (Exception ex) {
                    logger("error: " + ex.getMessage());
                }

            } else if ("kickDevice".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }
                d.leaveNetwork();

            } else if ("removePropertyListeners".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }
                d.removeAllPropertyUpdateHandlers();

            } else if ("updateProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                
                logger("Enter a property name: ");
                final String propertyId = scanner.nextLine();
                
                logger("Enter a value: ");
                final String writeValue = scanner.nextLine();

                Property property = new Property(propertyId, "na", writeValue);
                d.updateProperty(property).getFuture().whenComplete( (result, throwable) -> {
                    if (throwable != null) {
                        logger("error: " + throwable.getMessage());
                        logger("suggestion: ensure you are forming the correct JSON request, with property key:value pairs");
                    } else if (result != null) {
                        logger("result: " + result.getValue());    
                    }
                });

            } else if ("on".equalsIgnoreCase(input) || "off".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                
                final LightDevice light = (LightDevice) d;
                CompletableFuture<Property> f = null;
                if ("on".equalsIgnoreCase(input)) { f = light.on().getFuture(); }
                else if ("off".equalsIgnoreCase(input)) { f = light.off().getFuture(); }
                
                // print errors if they happen
                if (f != null) {
                    f.whenComplete( (result, throwable) -> {
                        if (throwable != null) {
                            logger("error: " + throwable.getMessage());
                        }
                    });
                }

            } else if ("moveToLevel".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                
                logger("Enter a level (as percent, 0-100): ");
                final String levelPercentageStr = scanner.nextLine();
                int levelPercentage = 0;
                try {
                    levelPercentage = Integer.parseInt(levelPercentageStr);
                } catch (NumberFormatException e) {
                    logger("invalid number");
                    continue;
                }
                
                final LightDevice light = (LightDevice) d;
                light.moveToLevel(levelPercentage);

            } else if ("readLevel".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }

                final LightDevice light = (LightDevice) d;
                try {
                    String level = light.readLevel().get(5, TimeUnit.SECONDS).getValue();
                    logger("Light Level for " + light.getID() + ": " + level);
                } catch (InterruptedException | ExecutionException | TimeoutException e) {}

            } else if ("sendZigBeeMessage".equalsIgnoreCase(input)) {
                logger("Enter Device ID:");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("device not found");
                    continue;
                }
                logger("enter json message:");
                String json = scanner.nextLine();
                try {
                    CompletableFuture<String> future = d.sendProtocolMessage(json);
                    String result = future.get(30, TimeUnit.SECONDS);
                    System.err.println("Result: " + result);
                } catch (InvalidInputException e) {
                    logger("Invalid JSON: "+ json);
                    e.printStackTrace();
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    e.printStackTrace();
                }
            } else if (("change mode".equalsIgnoreCase(input))
                    || ("change fan mode".equalsIgnoreCase(input))
                    || ("adjust setpoint".equalsIgnoreCase(input))
                    || ("read mode".equalsIgnoreCase(input))
                    || ("read fan mode".equalsIgnoreCase(input))
                    || ("read running mode".equalsIgnoreCase(input))
                    || ("read setpoint limit".equalsIgnoreCase(input)) 
                    || ("change setpoint limit".equalsIgnoreCase(input))) {
                
                // get the device
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null ) { logger("Device not found"); continue; }
                if (!(d instanceof ThermostatDevice)) { logger("device is not a thermostat"); continue; }

                // parse the command
                final ThermostatDevice device = (ThermostatDevice) d;
                PropertyCommandData command = null;
                CompletableFuture<Property> propertyFuture = null;
                if ("read mode".equalsIgnoreCase(input)) {
                    propertyFuture = device.readMode();
                } else if ("change mode".equalsIgnoreCase(input)) {
                    final StringBuilder modes = new StringBuilder();
                    modes.append("enter mode (");
                    modes.append(SystemMode.listModes());
                    modes.append("): ");
                    logger(modes.toString());
                    final String value = scanner.nextLine();
                    for (SystemMode systemMode : SystemMode.values()) {
                        if (systemMode.name().replace("_","").equalsIgnoreCase(value.replace(" ", ""))) {
                           command = device.changeMode(systemMode);
                           break;
                        }
                    }
                } else if ("adjust setpoint".equalsIgnoreCase(input)) {
                    Optional<CommandData> setpointCommand = Optional.empty();
                    final StringBuilder modes = new StringBuilder();
                    modes.append("enter mode (");
                    modes.append(Setpoint.listModes());
                    modes.append("): ");
                    logger(modes.toString());
                    final String modeInput = scanner.nextLine();
                    
                    logger("enter amount (from -127 to 127, where 1 = 0.1 degrees): ");
                    final String amountInput = scanner.nextLine();
                    for (Setpoint setpoint : Setpoint.values()) {
                        if (setpoint.name().equalsIgnoreCase(modeInput)) {
                           setpointCommand = device.adjustSetpoint(setpoint, Integer.parseInt(amountInput));
                        }
                    }
                    if (!setpointCommand.isPresent()) {
                        logger("command is null");
                        continue;
                    } else {
                        CommandData setPointCMD = setpointCommand.get();
                        logger("command cache result: " + setPointCMD.getCachedResult());
                        setPointCMD.getCommandResult().whenComplete((result, error) -> {
                            logger("Command: " + input + " finished");
                            if (error != null) {
                                error.printStackTrace();
                            } else {
                                logger("result: " + result);
                            }
                        });
                    }
                } else if ("change fan mode".equalsIgnoreCase(input)) {
                    final StringBuilder modes = new StringBuilder();
                    modes.append("enter mode (");
                    modes.append(FanMode.listModes());
                    modes.append("): ");
                    logger(modes.toString());
                    final String modeInput = scanner.nextLine();
                    for (FanMode fanMode : FanMode.values()) {
                        if (fanMode.name().equalsIgnoreCase(modeInput.toLowerCase())) {
                            command = device.changeFanMode(fanMode);
                            break;
                        }
                    }
                } else if ("read running mode".equalsIgnoreCase(input)) {
                    propertyFuture = device.readRunningMode();
                } else if ("read fan mode".equalsIgnoreCase(input)) {
                    propertyFuture = device.readFanMode();
                } else if ("read setpoint limit".equalsIgnoreCase(input)) {
                    final StringBuilder modes = new StringBuilder();
                    modes.append("enter mode (");
                    modes.append(Setpoint.listModes());
                    modes.append("): ");
                    logger(modes.toString());
                    final String modeInput = scanner.nextLine();
                    for (Setpoint setpoint : Setpoint.values()) {
                        if (setpoint.name().equalsIgnoreCase(modeInput)) {
                            propertyFuture = device.readSetpointLimit(setpoint);
                        }
                    }
                } else if ("change setpoint limit".equalsIgnoreCase(input)) {
                    final StringBuilder modes = new StringBuilder();
                    modes.append("enter mode (");
                    modes.append(Setpoint.listModes());
                    modes.append("): ");
                    logger(modes.toString());
                    final String modeInput = scanner.nextLine();
                    logger("enter amount, where 1 = 1 degrees celsius (i.e. 21.5 = 21.5 degrees Celsius): ");
                    final String amountInput = scanner.nextLine();
                    for (Setpoint setpoint : Setpoint.values()) {
                        if (setpoint.name().equalsIgnoreCase(modeInput)) {
                            command = device.setSetpointLimit(setpoint, Double.parseDouble(amountInput));
                        }
                    }
                }
                
                // handle the command results
                if (command == null && propertyFuture == null) {
                    logger("command is null");
                    continue;
                }
                if (command != null) {
                    Optional<Property> value = command.getCachedProperty();
                    if (value.isPresent()) {
                        logger("command cache result: " + value.get().getValue());
                    }
                    command.getFuture().whenComplete((result, error) -> {
                        logger("Command: " + input + " finished");
                        if (error != null) {
                            error.printStackTrace();
                        } else {
                            logger("result: " + result.getValue());
                        }
                    });
                } else {
                    propertyFuture.whenComplete((result, error) -> {
                        logger("Command: " + input + " finished");
                        if (error != null) {
                            error.printStackTrace();
                        } else {
                            logger("result: " + result.getValue());
                        }
                    });
                }
            } else {
                logger("Unknown command; try h for a list of commands");
            }
        }

        // clean up
        final boolean shutdown = gw.shutdown();
        if (!shutdown) {
            logger("Error shutting down gateway client, check logs");
        }
        scanner.close();

        // exit
        logger("Exited SampleApp");
    }
    
    private static final void logger(final String msg) {
        System.out.println(msg);
    }
    
    private static final void printCommands(final String[] commands) {
        logger("Available Commands: ");
        for (String s : commands) {
            logger(s);
        }
    }

    @Override
    public void onDeviceEvent(DeviceEvent event) {
        final Device d = event.getDevice();
        final DeviceEventStatus eventStatus = event.getStatus(); 
        
        logger("Device Event: " + eventStatus);
        logger("Device ID: " + d.getID());
        
        switch (eventStatus) {
        case DEVICE_ADDED:
            d.addPropertyUpdateHandler( (device, property) -> {
                logger("Property update triggered for device: " + device.getID());
                logger("Property: " + property.getName() + ", value: " + property.getValue() + ", type: " + property.getType());
            });
            break;
        default:
            break;
        }
    }
}
