package gatewayapi.sample.iaszone;

import java.util.Collection;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.mmbnetworks.gatewayapi.PropertyCommandData;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.gatewayapi.event.device.DeviceEvent;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventHandler;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventStatus;
import com.mmbnetworks.gatewayapi.exception.InvalidInputException;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.PropertyNames;
import com.mmbnetworks.rotarrandevicemodel.zigbee.property.IASCIEAddressProperty;
import com.mmbnetworks.rotarrandevicemodel.zigbee.property.ZoneStatusProperty;

public class SampleDeviceEventHandler implements DeviceEventHandler {

    private final SamplePropertyHandler propertyHandler;
    private final SampleProtocolPassthroughHandler protocolHandler;
    
    public SampleDeviceEventHandler() {
        this.propertyHandler = new SamplePropertyHandler();
        this.protocolHandler = new SampleProtocolPassthroughHandler();
    }

    @Override
    public void onDeviceEvent(DeviceEvent event) {
        // This method gets called when a device is added to, or removed from, the network.
        // By default, client callbacks are called from a single thread managed by GAPI.
        // Therefore, avoid writing blocking code in this method, or other client callbacks
        // (such as PropertyUpdateHandlers, or ProtocolHandlers) may be blocked.
        
        final DeviceEventStatus eventStatus = event.getStatus();
        final Device device = event.getDevice();
        final String deviceId = device.getID();
        final String deviceType = device.getDeviceType().getType();
        
        System.out.println("Device Event: " + eventStatus);
        System.out.println("Device Id: " + deviceId);
        System.out.println("Device Type: " + deviceType);
        
        switch (eventStatus) {
        
        case DEVICE_ADDED:
            // Register a property handler, to be notified when a Property on the Device changes.
            device.addPropertyUpdateHandler(propertyHandler);
            
            // Register a protocol passthru handler, to be notified about low-level
            // protocol-specific messages (i.e. zigbee messages from specific clusters)
            device.addProtocolHandler(protocolHandler);

            // Discover the device; this must be done before you can interact with device properties,
            // so it's good to do it when the system first learns about the device on the network.
            // This is a non-blocking call.
            CompletableFuture<Collection<Property>> properties = device.getProperties();
            
            // Do some extra (asynchronous) handling on the properties, once they've been discovered.
            properties.thenAccept( collection -> {
                // This code will be called from a GAPI thread.
                // Again, we shouldn't write blocking code in this section.
                for (Property p : collection) {
                    printDiscoveredProperty(p);
                    handleCieAddressRegistration(device, p);
                    handleZoneStatusBinding(device, p);
                }
            });

            break;

        case DEVICE_REMOVED:
            device.removeAllProtocolHandlers();
            device.removeAllPropertyUpdateHandlers();
            break;

        default:
            break;
        }
    }
    
    private static void printDiscoveredProperty(Property p) {
        StringBuilder sb = new StringBuilder();
        sb.append("Discovered Property: ").append(p.getName())
          .append(", with value: ").append(p.getValue())
          .append(", of type: ").append(p.getType());
        System.out.println(sb.toString());
    }
    
    private static void handleCieAddressRegistration(final Device device, final Property property) {
        // IAS Zone devices require that the address of the gateway be written to one of its attributes
        // before the device will function, so we do that here
        if (property.getName().equalsIgnoreCase(IASCIEAddressProperty.NAME)) {
            final String gatewayEui = getConnectedGatewayEui(device);
            if ( !gatewayEui.isEmpty() && !property.getValue().equalsIgnoreCase(gatewayEui) ) {
                writeCieAddress(device, gatewayEui);
            }
        }
    }
    
    private static void handleZoneStatusBinding(final Device device, final Property property) {
        if (property.getName().equalsIgnoreCase(ZoneStatusProperty.NAME)) {
            // do a binding on the ZoneStatus property, so our PropertyUpdateHandler will trigger
            // when the ZoneStatus changes.
            try {
                device.bindDevice(device.getConnectedGatewayId().get(), ZoneStatusProperty.NAME)
                      .get().getCommandResult()
                      .thenAccept(resultStr -> System.out.println("binding result: " + resultStr));

            } catch (InvalidInputException e) {
                System.out.println("error performing zone status binding");
            }   
        }
    }

    private static String getConnectedGatewayEui(final Device device) {
        // Every Device stores protocol-specific attributes (like a zigbee device's EUI), in a Property object
        // called "ProtocolProperty".
        // This code demonstrates how to grab the ProtocolProperty of the gateway that a device is connected to,
        // and then grab the EUI out of it.
        // It will return the EUI of the connected gateway if it exists, or an empty string if an error occurred.
        String gatewayEui = "";
        try {
            Optional<Property> o = SampleIasZoneApp.gw.getDevice(device.getConnectedGatewayId().get()).getCachedProperty(PropertyNames.PROTOCOL_PROPERTY_NAME);
            if (o.isPresent()) {
                final JsonParser parser = new JsonParser();
                final JsonObject jsonProperty = parser.parse(o.get().getValue()).getAsJsonObject();
                gatewayEui = jsonProperty.get("eui").getAsString();
            }
        } catch (Exception ex) {
            gatewayEui = "";    
        }
        return gatewayEui;
    }

    private static PropertyCommandData writeCieAddress(final Device device, final String eui) {
        // this is a non-blocking call to update the named property on a remote device
        System.out.println("Writing CIE Address: " + eui);
        return device.updateProperty(new Property(IASCIEAddressProperty.NAME, IASCIEAddressProperty.TYPE, eui));
    }

}
