package gatewayapi.sample.iaszone;

import java.util.Collection;
import java.util.Optional;
import java.util.Scanner;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.mmbnetworks.gatewayapi.ConnectionInfo;
import com.mmbnetworks.gatewayapi.ConnectionInfo.ConnectionType;
import com.mmbnetworks.gatewayapi.GatewayClient;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.LightDevice;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.gatewayapi.exception.GatewayConnectionException;
import com.mmbnetworks.gatewayapi.exception.InvalidInputException;

import ch.qos.logback.classic.Level;

/**
 * This sample application demonstrates how one might use the GatewayClient 
 * to handle events from an IAS Zone device. 
 */
public class SampleIasZoneApp {
    
    static GatewayClient gw = null;

    public static void main(String[] args) {

        Scanner scanner = new Scanner(System.in, "UTF-8");

        // check args
        if (args.length != 1) {
            while (true) {
                logger("Enter a serial port (i.e. /dev/ttyUSB0), or 'q' to exit:");
                final String input = scanner.nextLine();

                if ("q".equalsIgnoreCase(input)) {
                    logger("Quitting...");
                    scanner.close();
                    return;

                } else {
                    try {
                        gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, input), new SampleDeviceEventHandler());
                        break;
                    } catch (GatewayConnectionException e) { 
                        logger("Could not connect or create GatewayClient, try again");
                    }
                }
            }
        } else {
            final String devicePort = args[0];
            try {
                gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, devicePort), new SampleDeviceEventHandler());
            } catch (GatewayConnectionException e) { 
                logger("Could not connect or create GatewayClient with port: " + devicePort);
                scanner.close();
                return;
            }
        }

        // print available commands
        final String[] commands = {
                "==== General Commands ====",
                " h - help",
                " toggleLog - turn logging on/off",
                " v - get version",
                " n - network status",
                " f - form network",
                " p - permit join (30 sec)",
                " c - close network permit join window",
                " l - leave network",
                " q - quit",
                " d - list devices and their device ids",
                "==== Device Commands ====",
                " getDeviceType - get a list of device ids of a certain type",
                " getLights - get all the lighting device ids",
                " getProperties - get properties for a selected device id. Will perform a network request",
                " getCachedProperties - get last known properties for a selected device id. Will be empty if there are no last cached properties",
                " getProtocolProperties - gets the properties of the device in the format of the underlying protocol (i.e. zigbee). Will perform a network request",
                " getCachedProtocolProperties - gets the last known properties of the device in the format of the underlying protocol (i.e. zigbee). Will be empty if there are no last cached properties",
                " getProperty - get a specific property. Will perform a network request",
                " getCachedProperty - get the last known value of a specific property. Will be empty if there is no last known value. Will throw exception if property doesn't exist",
                " updateProperty - update a specific property for a selected device id",
                " bind - bind a source to a destination",
                " unbind - unbind a source to a destination",
                " enableReporting - enable reporting on properties using default parameters",
                " kickDevice - kick a remote device off the network",
                " removePropertyListeners - remove property update listeners for the selected device",
                "==== Light Device Commands ====",
                " on",
                " off",
                " moveToLevel - change the dimmable light level",
                " readLevel - read the dimmable light level",
                "========================",
        };
        printCommands(commands);
        final ch.qos.logback.classic.Logger rootLogger = (ch.qos.logback.classic.Logger) org.slf4j.LoggerFactory.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME);
        final Level originalLogLevel = rootLogger.getLevel();
        Level currentLogLevel = originalLogLevel;
        rootLogger.setLevel(Level.OFF);
        currentLogLevel = Level.OFF;

        // use the main thread as a CLI
        while (true) {
            logger("Enter a command: ");
            final String input = scanner.nextLine();
            
            if ("h".equalsIgnoreCase(input)) {
                printCommands(commands);

            } else if ("toggleLog".equalsIgnoreCase(input)) {
                if (currentLogLevel != Level.OFF) { 
                    rootLogger.setLevel(Level.OFF);
                    currentLogLevel = Level.OFF;
                } else {
                    rootLogger.setLevel(originalLogLevel);
                    currentLogLevel = originalLogLevel;
                }

            } else if ("v".equalsIgnoreCase(input)) {
                logger("API Version = " + GatewayClient.getApiVersion());

            } else if ("q".equalsIgnoreCase(input)) {
                break;

            } else if ("n".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.getNetworkStatus(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });   
                }

            } else if ("f".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.createNetwork(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("p".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.scanForDevices(c, 30).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("c".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.scanForDevices(c, 0).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("l".equalsIgnoreCase(input)) {
                for (ConnectionInfo c: gw.getConnectionInfo()) {
                    gw.dissolveNetwork(c).thenAccept( status -> {
                        logger("Status for connection " + c.getValue() + " is " + status);
                    });
                }

            } else if ("d".equalsIgnoreCase(input)) {
                final Collection<Device> devices = gw.getDevices();
                if (devices.size() == 0) {
                    logger("No devices found on the network");

                } else {
                    int i = 1;
                    for (Device d : devices) {
                        logger(i++ + ") Device Id: " + d.getID());
                    }
                }

            } else if ("getDeviceType".equalsIgnoreCase(input)) {
                logger("Enter a device type (i.e. light, thermostat, etc): ");
                final String typeStr = scanner.nextLine();
                final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().getType().equalsIgnoreCase(typeStr) );
                int i = 1;
                for (Device d : devices) {
                    logger(i++ + ") " + typeStr + " device id: " + d.getID());
                }

            } else if ("getLights".equalsIgnoreCase(input)) {
                final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().equals(LightDevice.DEVICE_TYPE));
                int i = 1;
                for (Device d : devices) {
                    logger(i++ + ") LightDevice Id: " + d.getID());
                }

            } else if ("getCachedProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }

                final Collection<Property> properties = d.getCachedProperties();
                if (properties.isEmpty()) {
                    logger("No properties found");
                } else {
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }   
                }

            } else if ("getProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                final CompletableFuture<Collection<Property>> result = d.getProperties();
                logger("Retrieving properties...");
                try {
                    Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated properties");
                    e.printStackTrace();
                }
            
            } else if ("getCachedProtocolProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }

                final Collection<Property> properties = d.getCachedProtocolProperties();
                if (properties.isEmpty()) {
                    logger("No properties found");
                } else {
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                }
                
            } else if ("getProtocolProperties".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                final CompletableFuture<Collection<Property>> result = d.getProtocolProperties();
                logger("Retrieving protocol properties...");
                try {
                    Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                    for (Property p : properties) {
                        logger("name: " + p.getName() + ", value: " + p.getValue() + ", type: " + p.getType());
                    }
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated properties: " + e.getMessage());
                    e.printStackTrace();
                }
            
            } else if ("getProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                
                logger("Enter a property name: ");
                final String propertyName = scanner.nextLine();
                
                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                final CompletableFuture<Property> p = device.getProperty(propertyName);
                try {
                    logger("Retrieving property \'" + propertyName +"\'");
                    final Property fProp = p.get(30, TimeUnit.SECONDS);
                    logger("name: " + fProp.getName() + ", value: " + fProp.getValue());
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    logger("Error, failed to retrieve updated property: " + e.getMessage());
                }

            } else if ("getCachedProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();

                logger("Enter a property name: ");
                final String propertyName = scanner.nextLine();

                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                final Optional<Property> pOpt = device.getCachedProperty(propertyName);
                if (!pOpt.isPresent()) {
                    logger("Property \'" + propertyName + "\' not found");
                    continue;
                }
                Property fProp = pOpt.get();
                logger("name: " + fProp.getName() + ", value: " + fProp.getValue());
            
            } else if ("enableReporting".equalsIgnoreCase(input)) {
                logger("Enter a device id:");
                String idStr = scanner.nextLine();
                final Device device = gw.getDevice(idStr);
                if (device == null) { logger("device not found"); continue; }

                try {
                    final CompletableFuture<String> f = device.enableDefaultReporting();
                    final String result = f.get(30, TimeUnit.SECONDS);
                    logger("Status for command: " + result);
                } catch (IllegalStateException | InterruptedException | ExecutionException | TimeoutException e) {
                    logger(e.getMessage());
                }
            
            } else if ("bind".equalsIgnoreCase(input)) {
                try {
                    logger("Enter a source id:");
                    final String source = scanner.nextLine();
                    final Device sourceDevice = gw.getDevice(source);
                    logger("Enter a destination id:");
                    final String destination = scanner.nextLine();
                    final Device destinationDevice = gw.getDevice(destination);
                    logger("Enter a Property name to bind:");
                    final String property = scanner.nextLine();
                    sourceDevice.bindDevice(destinationDevice.getID(), property).get()
                        .getCommandResult()
                        .whenComplete((result, error) -> {
                            if (error != null) {
                                logger("error calling bind: " + error.getMessage());
                            }
                            logger("bind device result: " + result);
                        });

                } catch (Exception ex) {
                    logger("error: " + ex.getMessage());
                }

            } else if ("unbind".equalsIgnoreCase(input)) {
                try {
                    logger("Enter a source id:");
                    final String source = scanner.nextLine();
                    final Device sourceDevice = gw.getDevice(source);
                    logger("Enter a destination id:");
                    final String destination = scanner.nextLine();
                    final Device destinationDevice = gw.getDevice(destination);
                    logger("Enter a Property name to unbind:");
                    final String property = scanner.nextLine();
                    sourceDevice.unbindDevice(destinationDevice.getID(), property).get()
                        .getCommandResult()
                        .whenComplete((result, error) -> {
                            if (error != null) {
                                logger("error calling unbind: " + error.getMessage());
                            }
                            logger("unbind device result: " + result);
                        });

                } catch (Exception ex) {
                    logger("error: " + ex.getMessage());
                }

            } else if ("kickDevice".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }
                d.leaveNetwork();

            } else if ("removePropertyListeners".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("Device not found"); continue; }
                d.removeAllPropertyUpdateHandlers();

            } else if ("updateProperty".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("Device not found");
                    continue;
                }
                
                logger("Enter a property name: ");
                final String propertyId = scanner.nextLine();
                
                logger("Enter a value: ");
                final String writeValue = scanner.nextLine();

                Property property = new Property(propertyId, "na", writeValue);
                d.updateProperty(property).getFuture().whenComplete( (result, throwable) -> {
                    if (throwable != null) {
                        logger("error: " + throwable.getMessage());
                        logger("suggestion: ensure you are forming the correct JSON request, with property key:value pairs");
                    } else if (result != null) {
                        logger("result: " + result.getValue());    
                    }
                });

            } else if ("on".equalsIgnoreCase(input) || "off".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                
                final LightDevice light = (LightDevice) d;
                CompletableFuture<Property> f = null;
                if ("on".equalsIgnoreCase(input)) { f = light.on().getFuture(); }
                else if ("off".equalsIgnoreCase(input)) { f = light.off().getFuture(); }
                
                // print errors if they happen
                if (f != null) {
                    f.whenComplete( (result, throwable) -> {
                        if (throwable != null) {
                            logger("error: " + throwable.getMessage());
                        }
                    });
                }

            } else if ("moveToLevel".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                
                logger("Enter a level (as percent, 0-100): ");
                final String levelPercentageStr = scanner.nextLine();
                int levelPercentage = 0;
                try {
                    levelPercentage = Integer.parseInt(levelPercentageStr);
                } catch (NumberFormatException e) {
                    logger("invalid number");
                    continue;
                }
                
                final LightDevice light = (LightDevice) d;
                light.moveToLevel(levelPercentage);

            } else if ("readLevel".equalsIgnoreCase(input)) {
                logger("Enter a device id: ");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) { logger("device not found"); continue; }
                if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }

                final LightDevice light = (LightDevice) d;
                try {
                    String level = light.readLevel().get(5, TimeUnit.SECONDS).getValue();
                    logger("Light Level for " + light.getID() + ": " + level);
                } catch (InterruptedException | ExecutionException | TimeoutException e) {}

            } else if ("sendZigBeeMessage".equalsIgnoreCase(input)) {
                logger("Enter Device ID:");
                final String idStr = scanner.nextLine();
                final Device d = gw.getDevice(idStr);
                if (d == null) {
                    logger("device not found");
                    continue;
                }
                logger("enter json message:");
                String json = scanner.nextLine();
                try {
                    CompletableFuture<String> future = d.sendProtocolMessage(json);
                    String result = future.get(30, TimeUnit.SECONDS);
                    System.err.println("Result: " + result);
                } catch (InvalidInputException e) {
                    logger("Invalid JSON: "+ json);
                    e.printStackTrace();
                } catch (InterruptedException | ExecutionException | TimeoutException e) {
                    e.printStackTrace();
                }

            } else {
                logger("Unknown command; try h for a list of commands");
            }
        }

        // clean up
        final boolean shutdown = gw.shutdown();
        if (!shutdown) {
            logger("Error shutting down gateway client, check logs");
        }
        scanner.close();

        // exit
        logger("Exited SampleApp");
    }
    
    private static final void logger(final String msg) {
        System.out.println(msg);
    }
    
    private static final void printCommands(final String[] commands) {
        logger("Available Commands: ");
        for (String s : commands) {
            logger(s);
        }
    }

}
